/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <kozo.h>

#include "bkl-db.h"
#include "bkl-item-audio.h"
#include "bkl-item-image.h"
#include "bkl-item-video.h"

enum {
    FIELD_TYPE,
    FIELD_FILE,
    FIELD_AUDIO,
    FIELD_IMAGE,
    FIELD_VIDEO,
    FIELD_EXTENDED,
    LAST_FIELD
};

static const char *fieldnames[LAST_FIELD] = {
    "type", "file", "audio", "image", "video", "extended"
};
int fieldids[LAST_FIELD] = {0, };

static gboolean
register_fields (KozoDB  *db,
                 GError **error)
{
    int i, id;

    for (i = 0; i < LAST_FIELD; i++) {
        id = kozo_db_register_field (db, fieldnames[i], error);
        if (id < 0) {
            return FALSE;
        }

        fieldids[i] = id;
    }

    return TRUE;
}

KozoDB *
bkl_db_get (const char *db_name,
            GError    **error)
{
    KozoDB *db;

    db = kozo_db_get (db_name, BICKLEY_DB_VERSION, error);
    if (db == NULL) {
        return NULL;
    }

    if (register_fields (db, error)) {
        kozo_db_flush (db);
        return db;
    } else {
        kozo_db_unref (db);
        return NULL;
    }
}

KozoDB *
bkl_db_get_for_path (const char *path,
                     const char *db_name,
                     GError    **error)
{
    KozoDB *db;

    db = kozo_db_get_for_path (path, db_name,
                               BICKLEY_DB_VERSION, error);
    if (db == NULL) {
        return NULL;
    }

    if (register_fields (db, error)) {
        kozo_db_flush (db);
        return db;
    } else {
        kozo_db_unref (db);
        return NULL;
    }
}

KozoDB *
bkl_db_get_for_name (const char *filename,
                     const char *db_name,
                     GError    **error)
{
    KozoDB *db;

    db = kozo_db_get_for_name (filename, db_name, BICKLEY_DB_VERSION, error);
    if (db == NULL) {
        return NULL;
    }

    if (register_fields (db, error)) {
        kozo_db_flush (db);
        return db;
    } else {
        kozo_db_unref (db);
        return NULL;
    }
}

gboolean
bkl_db_add_item (KozoDB     *db,
                 const char *key,
                 BklItem    *item,
                 GError    **error)
{
    GSList *fields;
    gboolean ret;

    fields = bkl_item_get_fields (item);
    ret = kozo_db_add (db, key, fields, error);
    kozo_field_list_free (fields);

    return ret;
}

gboolean
bkl_db_replace_item (KozoDB     *db,
                     const char *key,
                     BklItem    *item,
                     GError    **error)
{
    GSList *fields;
    gboolean ret;

    fields = bkl_item_get_fields (item);

    ret = kozo_db_set (db, key, fields, error);
    kozo_field_list_free (fields);

    return ret;
}

gboolean
bkl_db_remove (KozoDB     *db,
               const char *key,
               GError    **error)
{
    return kozo_db_remove (db, key, error);
}

static BklItem *
item_from_entry (KozoEntry *entry)
{
    BklItem *item;
    BklItemType type;
    KozoField *field;

    field = kozo_entry_get_field (entry, FIELD_TYPE);
    type = kozo_field_get_value_int (field, 0);

    switch (type) {
    case BKL_ITEM_TYPE_AUDIO:
        field = kozo_entry_get_field (entry, FIELD_AUDIO);
        item = (BklItem *) bkl_item_audio_new_from_field (field);
        break;

    case BKL_ITEM_TYPE_IMAGE:
        field = kozo_entry_get_field (entry, FIELD_IMAGE);
        item = (BklItem *) bkl_item_image_new_from_field (field);
        break;

    case BKL_ITEM_TYPE_VIDEO:
        field = kozo_entry_get_field (entry, FIELD_VIDEO);
        item = (BklItem *) bkl_item_video_new_from_field (field);
        break;

    default:
        return NULL;
    }

    field = kozo_entry_get_field (entry, FIELD_FILE);
    bkl_item_set_from_field (item, field);

    field = kozo_entry_get_field (entry, FIELD_EXTENDED);
    bkl_item_extended_set_from_field ((BklItemExtended *) item, field);

    return item;
}

BklItem *
bkl_db_get_item (KozoDB     *db,
                 const char *key,
                 GError    **error)
{
    KozoEntry *entry;

    entry = kozo_db_lookup (db, key, NULL, error);
    if (entry == NULL) {
        return NULL;
    }

    return item_from_entry (entry);
}

static void
build_item_list (KozoDB     *db,
                 const char *key,
                 KozoEntry  *entry,
                 gpointer    userdata)
{
    GPtrArray *array = userdata;
    BklItem *item;

    item = item_from_entry (entry);
    if (item != NULL) {
        g_ptr_array_add (array, item);
    }
}

GPtrArray *
bkl_db_get_items (KozoDB  *db,
                  GError **error)
{
    GPtrArray *array;

    array = g_ptr_array_new ();
    kozo_db_foreach (db, build_item_list, array);

    return array;
}
