/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>
#include <glib/gthread.h>

#include "kozo-query-private.h"
#include "kozo-db-private.h"
#include "kozo-db-query.h"
#include "kozo-expr-private.h"
#include "kozo-pack.h"

/* Tweakable parameters */

/* Iterate DB in 1MB chunks */
#define ITERATION_CHUNK_SIZE 1048576

typedef struct {
        KozoQuery parent;

        gboolean (* relevance_func) (KozoQuery *query,
                                     const TDB_DATA key,
                                     const TDB_DATA data);
} KozoDBQuery;

static void
add_result (KozoQuery       *query,
            KozoEntryArray  *array,
            KozoPackContext *context,
            const TDB_DATA   key,
            const TDB_DATA   data)
{
        kozo_pack_start (context);
        kozo_db_harvest_entry (key, data, context, query->field_ids);
        context->dest =
                obstack_alloc (&array->obstack, context->length);
        kozo_pack_finish (context);

        kozo_entry_array_append (array,
                                 context->dest, context->length);
}

static gboolean
and_relevance_func (KozoQuery     *query,
                    const TDB_DATA key,
                    const TDB_DATA data)
{
        GSList *l;

        for (l = query->exprs; l; l = l->next)
                if (!kozo_expr_eval (l->data, key, data))
                        return FALSE;

        return TRUE;
}

static gboolean
or_relevance_func (KozoQuery     *query,
                   const TDB_DATA key,
                   const TDB_DATA data)
{
        GSList *l;

        for (l = query->exprs; l; l = l->next)
                if (kozo_expr_eval (l->data, key, data))
                        return TRUE;

        return FALSE;
}

static gboolean
always_relevance_func (KozoQuery *query,
                       const TDB_DATA key,
                       const TDB_DATA data)
{
        return TRUE;
}

static gboolean
kozo_db_query_run (KozoQuery      *query,
                   KozoEntryArray *results,
                   GError        **error)
{
        KozoDBQuery *db_query;
        TDB_CONTEXT *db;
        TDB_DATA key, data, newkey;
        KozoPackContext context;

        db_query = (KozoDBQuery *) query;

        if (query->n_exprs > 0) {
                switch (query->op) {
                case KOZO_OPERATOR_AND:
                        db_query->relevance_func = and_relevance_func;
                        break;
                case KOZO_OPERATOR_OR:
                        db_query->relevance_func = or_relevance_func;
                        break;
                }
        } else
                db_query->relevance_func = always_relevance_func;

        db = kozo_db_get_db (query->db);
        g_return_val_if_fail (db != NULL, FALSE);

        kozo_pack_context_init (&context, FALSE,
                                query->n_field_ids, sizeof (guint32));

        key = tdb_firstkey (db);
        while (key.dptr) {
                data = tdb_fetch (db, key);

                if (db_query->relevance_func (query, key, data))
                        add_result (query, results, &context, key, data);

                free (data.dptr);
                newkey = tdb_nextkey (db, key);
                free (key.dptr);
                key = newkey;
        }

        return TRUE;
}

static gboolean
kozo_db_query_notify (KozoQuery      *query,
                      KozoEntryArray *queue,
                      const TDB_DATA  key,
                      const TDB_DATA  data,
                      guint8        **dest,
                      guint32        *length)
{
        KozoDBQuery *db_query;
        KozoPackContext context;

        db_query = (KozoDBQuery *) query;

        if (!db_query->relevance_func (query, key, data))
                return FALSE;

        kozo_pack_context_init (&context, FALSE,
                                query->n_field_ids, sizeof (guint32));
        kozo_pack_start (&context);
        kozo_db_harvest_entry (key, data, &context, query->field_ids);
        context.dest =
                obstack_alloc (&queue->obstack, context.length);
        kozo_pack_finish (&context);

        *dest = context.dest;
        *length = context.length;

        return TRUE;
}

KozoQuery *
kozo_db_query_new (KozoDB *db)
{
        KozoDBQuery *db_query;
        KozoQuery *query;

        db_query = g_new (KozoDBQuery, 1);
        query = (KozoQuery *) db_query;

        kozo_query_set_defaults (query, db);

        query->run = kozo_db_query_run;
        query->notify = kozo_db_query_notify;

        return query;
}
