/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>

#include "kozo-expr.h"
#include "kozo-expr-private.h"
#include "kozo-query.h"
#include "kozo-entry.h"
#include "kozo-entry-private.h"
#include "kozo-utf8.h"

/* Tweakable parameters */

/* Amount of expr structures per memory block */
#define EXPR_BLOCK_SIZE 16

static GMutex *expr_mem_lock;
static GMemChunk *expr_mem_chunk;
static GTrashStack *free_exprs = NULL;

typedef enum {
        KOZO_VALUE_STRING,
        KOZO_VALUE_INT,
        KOZO_VALUE_TIME
} KozoValueType;

struct _KozoExpr {
        KozoExprType type;
        int field_id;
        KozoValueType cnst_type;
        union {
               gunichar *str;
               gint32 i;
               time_t time;
        } cnst;

        gboolean (* expr_func) (KozoExpr      *expr,
                                const TDB_DATA key,
                                const TDB_DATA data);

        guint ref_count;
};

G_INLINE_FUNC gboolean
do_string_expr (KozoExpr *expr, const char *str)
{
        switch (expr->type) {
        case KOZO_EXPR_CONTAINS:
                return (kozo_utf8_strstrcasestrip
                                        (str, expr->cnst.str) != NULL);
        case KOZO_EXPR_DOES_NOT_CONTAIN:
                return (kozo_utf8_strstrcasestrip
                                        (str, expr->cnst.str) == NULL);
        case KOZO_EXPR_EQ:
                return (kozo_utf8_streqcasestrip
                                        (str, expr->cnst.str));
        case KOZO_EXPR_NE:
                return (!kozo_utf8_streqcasestrip
                                        (str, expr->cnst.str));
        default:
                g_warning ("Invalid expression");
                return FALSE;
        }
}

G_INLINE_FUNC gboolean
do_int_expr (KozoExpr *expr, gint32 i)
{
        switch (expr->type) {
        case KOZO_EXPR_EQ:
                return (i == expr->cnst.i);
        case KOZO_EXPR_NE:
                return (i != expr->cnst.i);
        case KOZO_EXPR_GE:
                return (i >= expr->cnst.i);
        case KOZO_EXPR_GT:
                return (i > expr->cnst.i);
        case KOZO_EXPR_LE:
                return (i <= expr->cnst.i);
        case KOZO_EXPR_LT:
                return (i < expr->cnst.i);
        default:
                g_warning ("Invalid expression");
                return FALSE;
        }
}

G_INLINE_FUNC gboolean
do_time_expr (KozoExpr *expr, time_t time)
{
        switch (expr->type) {
        case KOZO_EXPR_EQ:
                return (time == expr->cnst.time);
        case KOZO_EXPR_NE:
                return (time != expr->cnst.time);
        case KOZO_EXPR_GE:
                return (time >= expr->cnst.time);
        case KOZO_EXPR_GT:
                return (time > expr->cnst.time);
        case KOZO_EXPR_LE:
                return (time <= expr->cnst.time);
        case KOZO_EXPR_LT:
                return (time < expr->cnst.time);
        default:
                g_warning ("Invalid expression");
                return FALSE;
        }
}

static gboolean
do_key_expr_string (KozoExpr      *expr,
                    const TDB_DATA key,
                    const TDB_DATA data)
{
        return do_string_expr (expr, (const char *) key.dptr);
}

static gboolean
do_field_expr_string_and (KozoExpr      *expr,
                          const TDB_DATA key,
                          const TDB_DATA data)
{
        KozoEntry entry;
        KozoField *field;
        guint32 count, j;
        const char *str;
        gboolean ret;

        ret = TRUE;

        entry.data = data.dptr;
        entry.length = data.dsize;

        field = kozo_entry_get_field (&entry, expr->field_id);
        count = kozo_field_get_value_count (field);

        for (j = 0; j < count; j++) {
                str = kozo_field_get_value_string (field, j);

                if (!do_string_expr (expr, str)) {
                        ret = FALSE;
                        break;
                }
        }

        kozo_field_free (field);

        return ret;
}

static gboolean
do_field_expr_string_or (KozoExpr      *expr,
                         const TDB_DATA key,
                         const TDB_DATA data)
{
        KozoEntry entry;
        KozoField *field;
        guint32 count, j;
        const char *str;
        gboolean ret;

        ret = FALSE;

        entry.data = data.dptr;
        entry.length = data.dsize;

        field = kozo_entry_get_field (&entry, expr->field_id);
        count = kozo_field_get_value_count (field);

        for (j = 0; j < count; j++) {
                str = kozo_field_get_value_string (field, j);

                if (do_string_expr (expr, str)) {
                        ret = TRUE;
                        break;
                }
        }

        kozo_field_free (field);

        return ret;
}

static gboolean
do_field_expr_string (KozoExpr      *expr,
                      const TDB_DATA key,
                      const TDB_DATA data)
{
        KozoEntry entry;
        KozoField *field;
        guint32 count;
        gboolean ret;

        ret = FALSE;

        entry.data = data.dptr;
        entry.length = data.dsize;

        field = kozo_entry_get_field (&entry, expr->field_id);
        count = kozo_field_get_value_count (field);

        if (count == 1) {
                ret = do_string_expr (expr,
                                      kozo_field_get_value_string (field, 0));
        }

        kozo_field_free (field);

        return ret;
}

static gboolean
do_field_expr_int (KozoExpr      *expr,
                   const TDB_DATA key,
                   const TDB_DATA data)
{
        KozoEntry entry;
        KozoField *field;
        guint32 count;
        gboolean ret;

        ret = FALSE;

        entry.data = data.dptr;
        entry.length = data.dsize;

        field = kozo_entry_get_field (&entry, expr->field_id);
        count = kozo_field_get_value_count (field);

        if (count == 1) {
                ret = do_int_expr (expr,
                                   kozo_field_get_value_int (field, 0));
        }

        kozo_field_free (field);

        return ret;
}

static gboolean
do_field_expr_time (KozoExpr      *expr,
                    const TDB_DATA key,
                    const TDB_DATA data)
{
        KozoEntry entry;
        KozoField *field;
        guint32 count;
        gboolean ret;

        ret = FALSE;

        entry.data = data.dptr;
        entry.length = data.dsize;

        field = kozo_entry_get_field (&entry, expr->field_id);
        count = kozo_field_get_value_count (field);

        if (count == 1) {
                ret = do_time_expr (expr,
                                    kozo_field_get_value_time (field, 0));
        }

        kozo_field_free (field);

        return ret;
}

KozoExpr *
kozo_expr_new_string (KozoExprType type,
                      int          field_id,
                      const char  *cnst)
{
        KozoExpr *expr;

        g_mutex_lock (expr_mem_lock);

        expr = g_trash_stack_pop (&free_exprs);
        if (!expr)
                expr = g_chunk_new (KozoExpr, expr_mem_chunk);

        g_mutex_unlock (expr_mem_lock);

        expr->type = type;
        expr->field_id = field_id;
        expr->cnst_type = KOZO_VALUE_STRING;
        expr->cnst.str = kozo_utf8_strcasestrip (cnst);
        expr->ref_count = 1;

        if (field_id >= 0) {
                switch (expr->type) {
                case KOZO_EXPR_CONTAINS:
                        expr->expr_func = do_field_expr_string_or;
                        break;
                case KOZO_EXPR_DOES_NOT_CONTAIN:
                        expr->expr_func = do_field_expr_string_and;
                        break;
                default:
                        expr->expr_func = do_field_expr_string;
                        break;
                }
        } else
                expr->expr_func = do_key_expr_string;

        return expr;
}

KozoExpr *
kozo_expr_new_int (KozoExprType type,
                   int          field_id,
                   gint32       cnst)
{
        KozoExpr *expr;

        g_return_val_if_fail (field_id >= 0, NULL);

        g_mutex_lock (expr_mem_lock);

        expr = g_trash_stack_pop (&free_exprs);
        if (!expr)
                expr = g_chunk_new (KozoExpr, expr_mem_chunk);

        g_mutex_unlock (expr_mem_lock);
        
        expr->type = type;
        expr->field_id = field_id;
        expr->cnst_type = KOZO_VALUE_INT;
        expr->cnst.i = cnst;
        expr->ref_count = 1;

        expr->expr_func = do_field_expr_int;

        return expr;
}

KozoExpr *
kozo_expr_new_time (KozoExprType type,
                    int          field_id,
                    time_t       cnst)
{
        KozoExpr *expr;

        g_return_val_if_fail (field_id >= 0, NULL);

        g_mutex_lock (expr_mem_lock);

        expr = g_trash_stack_pop (&free_exprs);
        if (!expr)
                expr = g_chunk_new (KozoExpr, expr_mem_chunk);

        g_mutex_unlock (expr_mem_lock);
        
        expr->type = type;
        expr->field_id = field_id;
        expr->cnst_type = KOZO_VALUE_TIME;
        expr->cnst.time = cnst;
        expr->ref_count = 1;

        expr->expr_func = do_field_expr_time;

        return expr;
}

static void
kozo_expr_free (KozoExpr *expr)
{
        if (expr->cnst_type == KOZO_VALUE_STRING)
                g_free (expr->cnst.str);

        g_mutex_lock (expr_mem_lock);

        g_trash_stack_push (&free_exprs, expr);

        g_mutex_unlock (expr_mem_lock);
}

KozoExpr *
kozo_expr_ref (KozoExpr *expr)
{
        expr->ref_count++;

        return expr;
}

void
kozo_expr_unref (KozoExpr *expr)
{
        expr->ref_count--;

        if (expr->ref_count == 0)
                kozo_expr_free (expr);
}

gboolean
kozo_expr_eval (KozoExpr      *expr,
                const TDB_DATA key,
                const TDB_DATA data)
{
        return expr->expr_func (expr, key, data);
}

void
kozo_expr_init (void)
{
        expr_mem_lock = g_mutex_new ();

        expr_mem_chunk = g_mem_chunk_create (KozoExpr,
                                             EXPR_BLOCK_SIZE,
                                             G_ALLOC_ONLY);
}

void
kozo_expr_shutdown (void)
{
        g_mem_chunk_destroy (expr_mem_chunk);

        g_mutex_free (expr_mem_lock);
}
