/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <glib/gthread.h>
#include <string.h>

#include "kozo-index-private.h"
#include "kozo-index-query.h"
#include "kozo-db-private.h"
#include "kozo-private.h"
#include "kozo-pack.h"

struct _KozoIndex {
        KozoDB *db;
        GSList *field_ids;
        char *id;

        TDB_CONTEXT *index;

        KozoPackContext *pack_context;

        int ref_count;
};

static GHashTable *indeces;
static GMutex *indeces_lock;

static GPrivate *current_index;

#define NOT_IMPLEMENTED (g_print ("%s not implemented yet.\n", __FUNCTION__))

#if 0
static void
kozo_index_free (KozoIndex *index)
{
        char *fn;

        g_hash_table_remove (indeces, index->id);

        kozo_db_associate_lock (index->db);
        index->index->close (index->index, DB_NOSYNC);
        kozo_db_associate_unlock (index->db);

        /* Remove DB. We will rebuild when needed.
         * Otherwise the index would have to be open
         * at all times. */
        db_create (&index->index, NULL, 0);

        fn = g_build_filename (kozo_server_get_db_dir (), index->id, NULL);
        index->index->remove (index->index, fn, NULL, 0);
        g_free (fn);

        kozo_db_unref (index->db);
        g_slist_free (index->field_ids);
        g_free (index->id);

        g_free (index);
}

static int
get_secondary_key (DB        *db,
                   const DBT *pkey,
                   const DBT *pdata,
                   DBT       *skey)
{
        KozoIndex *index;

        index = (KozoIndex *) g_private_get (current_index);

        kozo_pack_start (index->pack_context);
        kozo_db_harvest_entry (pkey, pdata,
                               index->pack_context, index->field_ids);
        index->pack_context->dest =
                g_realloc (index->pack_context->dest,
                           index->pack_context->length);
        kozo_pack_finish (index->pack_context);

        memset (skey, 0, sizeof (DBT));

        skey->data = index->pack_context->dest;
        skey->size = index->pack_context->length;

        return 0;
}

static gboolean
build_index (KozoIndex *index,
             GError   **error)
{
        char *fn;
        KozoPackContext pack_context;
        TDB_CONTEXT *db;

        fn = g_build_filename (kozo_server_get_db_dir (), index->id, NULL);
        index->index = (fn, 0, 0, O_RDWR | O_TRUNC, 0644);
        g_free (fn);

        if (G_UNLIKELY (!index->index)) {
                /* kozo_db_set_g_error (error, err); */

                return FALSE;
        }

        kozo_pack_context_init (&pack_context, FALSE,
                                g_slist_length (index->field_ids),
                                sizeof (guint32));
        pack_context.dest = NULL;

        index->pack_context = &pack_context;

        /* Yay for lack of user_data :/ */
        g_private_set (current_index, index);

        db = kozo_db_get_db (index->db);

        kozo_db_reader_lock (index->db);
        kozo_db_associate_lock (index->db);

        /* This can take a while .. */
        db->associate (db, NULL, index->index, get_secondary_key, DB_CREATE);

        kozo_db_associate_unlock (index->db);
        kozo_db_reader_unlock (index->db);

        g_free (pack_context.dest);

        index->pack_context = NULL;

        return TRUE;
}

static KozoIndex *
kozo_index_new (KozoDB  *db,
                GSList  *field_ids,
                char    *id,
                GError **error)
{
        KozoIndex *index;

        index = g_new (KozoIndex, 1);

        index->db = kozo_db_ref (db);
        index->field_ids = g_slist_copy (field_ids);
        index->id = id;
        index->ref_count = 1;

        if (G_UNLIKELY (!build_index (index, error))) {
                kozo_index_free (index);
                return NULL;
        }

        return index;
}

static char *
calculate_id (KozoDB *db,
              GSList *field_ids)
{
        GString *str;
        GSList *l;
        
        str = g_string_new (kozo_db_get_name (db));
        g_string_append_c (str, '-');

        for (l = field_ids; l; l = l->next) {
                g_string_append_printf (str, "%d", GPOINTER_TO_INT (l->data));
                if (l->next)
                        g_string_append_c (str, '+');
        }

        return g_string_free (str, FALSE);
}
#endif

KozoIndex *
kozo_index_get (KozoDB  *db,
                GSList  *field_ids,
                GError **error)
{
#if 0
        KozoIndex *index;
        char *id;

        g_mutex_lock (indeces_lock);

        id = calculate_id (db, field_ids);
        
        index = g_hash_table_lookup (indeces, id);
        if (index) {
                index = kozo_index_ref (index);

                g_mutex_unlock (indeces_lock);

                g_free (id);

                return index;
        }

        index = kozo_index_new (db, field_ids, id, error);
        if (G_LIKELY (index))
                g_hash_table_insert (indeces, index->id, index);
        
        g_mutex_unlock (indeces_lock);

        return index;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

KozoIndex *
kozo_index_ref (KozoIndex *index)
{
#if 0
        g_atomic_int_inc (&index->ref_count);

        return index;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

void
kozo_index_unref (KozoIndex *index)
{
#if 0
        g_mutex_lock (indeces_lock);

        if (g_atomic_int_dec_and_test (&index->ref_count))
                kozo_index_free (index);

        g_mutex_unlock (indeces_lock);
#else
        NOT_IMPLEMENTED;
#endif
}

TDB_CONTEXT *
kozo_index_get_index (KozoIndex *index)
{
#if 0
        return index->index;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

const GSList *
kozo_index_get_field_ids (KozoIndex *index)
{
#if 0
        return (const GSList *) index->field_ids;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

KozoDB *
kozo_index_get_db (KozoIndex *index)
{
#if 0
        return index->db;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

KozoEntry *
kozo_index_lookup (KozoIndex    *index,
                   const guint8 *key,
                   guint32       key_length,
                   GSList       *field_ids,
                   GError      **error)
{
#if 0
        /* FIXME */
        return NULL;
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

KozoQuery *
kozo_index_new_query (KozoIndex *index)
{
#if 0
        return kozo_index_query_new (index);
#else
        NOT_IMPLEMENTED;
        return NULL;
#endif
}

void
kozo_index_init (void)
{
#if 0
        indeces = g_hash_table_new (g_str_hash, g_str_equal);

        indeces_lock = g_mutex_new ();

        current_index = g_private_new (NULL);
#else
        NOT_IMPLEMENTED;
#endif
}

void
kozo_index_shutdown (void)
{
#if 0
        g_mutex_free (indeces_lock);

        g_hash_table_destroy (indeces);
#else
        NOT_IMPLEMENTED;
#endif
}
