/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>
#include <glib/gthread.h>

#include "kozo-query-private.h"
#include "kozo-db-private.h"
#include "kozo-expr-private.h"
#include "kozo-index-private.h"
#include "kozo-pack.h"
#include "kozo-util.h"

/* Tweakable parameters */

/* Initial size of the resizing array. Should not be too large
 * or too small -- the array grows exponentially. */
#define INITIAL_RESULTS_SIZE 32

/* Emit signals after at least every 10 notifications */
#define QUEUE_THRESHOLD 10

/* Initial size of a notification queue. Should not be too large
 * or too small -- the array grows exponentially. */
#define INITIAL_QUEUE_SIZE 10

void
kozo_query_set_defaults (KozoQuery *query,
                         KozoDB    *db)
{
        query->db = kozo_db_ref (db);

        query->field_ids = NULL;
        query->n_field_ids = 0;

        query->op = KOZO_OPERATOR_OR;
        query->exprs = NULL;
        query->n_exprs = 0;

        query->entries_added = NULL;
        query->entries_changed = NULL;
        query->entries_removed = NULL;
        query->user_data = NULL;

        query->added_queue = kozo_entry_array_new (INITIAL_QUEUE_SIZE);
        query->changed_queue = kozo_entry_array_new (INITIAL_QUEUE_SIZE);
        query->removed_queue = kozo_entry_array_new (INITIAL_QUEUE_SIZE);
        query->queue_lock = g_mutex_new ();

        query->ref_count = 1;

        query->free = NULL;
}

static void
kozo_query_free (KozoQuery *query)
{
        if (query->free)
                query->free (query);

        kozo_db_remove_query (query->db, query);
        kozo_db_unref (query->db);

        while (query->exprs) {
                KozoExpr *expr;

                expr = (KozoExpr *) query->exprs->data;
                kozo_expr_unref (expr);

                query->exprs = g_slist_delete_link (query->exprs,
                                                    query->exprs);
        }

        g_slist_free (query->field_ids);

        kozo_entry_array_free (query->added_queue);
        kozo_entry_array_free (query->changed_queue);
        kozo_entry_array_free (query->removed_queue);
        g_mutex_free (query->queue_lock);

        g_free (query);
}

KozoQuery *
kozo_query_ref (KozoQuery *query)
{
        g_atomic_int_inc (&query->ref_count);

        return query;
}

void
kozo_query_unref (KozoQuery *query)
{
        KozoDB *db;

        /* We need to lock here as the DB contains a reference
         * to the query */
        db = kozo_db_ref (query->db);
        kozo_db_writer_lock (db);

        if (g_atomic_int_dec_and_test (&query->ref_count))
                kozo_query_free (query);

        kozo_db_writer_unlock (db);
        kozo_db_unref (db);
}

void
kozo_query_add_field (KozoQuery *query,
                      int        field_id)
{
        query->field_ids = g_slist_append (query->field_ids,
                                           GINT_TO_POINTER (field_id));
        query->n_field_ids++;
}

const GSList *
kozo_query_get_field_ids (KozoQuery *query)
{
        return query->field_ids;
}

void
kozo_query_set_op (KozoQuery   *query,
                   KozoOperator op)
{
        query->op = op;
}

KozoOperator
kozo_query_get_op (KozoQuery *query)
{
        return query->op;
}

void
kozo_query_add_expr (KozoQuery *query,
                     KozoExpr  *expr)
{
        query->exprs = g_slist_append (query->exprs,
                                       kozo_expr_ref (expr));
        query->n_exprs++;
}

const GSList *
kozo_query_get_exprs (KozoQuery *query)
{
        return query->exprs;
}

void
kozo_query_set_funcs (KozoQuery        *query,
                      KozoQueryCallback entries_added,
                      KozoQueryCallback entries_changed,
                      KozoQueryCallback entries_removed,
                      gpointer          user_data)
{
        query->entries_added   = entries_added;
        query->entries_changed = entries_changed;
        query->entries_removed = entries_removed;

        query->user_data = user_data;
}

gboolean
kozo_query_run (KozoQuery  *query,
                GError    **error)
{
        KozoEntryArray results;
        gboolean ret;

        kozo_db_reader_lock (query->db);
        kozo_db_add_query (query->db, query);

        kozo_entry_array_init (&results, INITIAL_RESULTS_SIZE);

        ret = query->run (query, &results, error);

        kozo_db_reader_unlock (query->db);

        if (query->entries_added) {
                query->entries_added (query,
                                      results.array,
                                      results.len,
                                      query->user_data);
        }

        kozo_entry_array_destroy (&results);

        return ret;
}

static void
flush_queue (KozoQuery          *query,
             KozoQueryNotifyType type)
{
        KozoEntryArray **queue, *old_queue;
        KozoQueryCallback func;

        g_mutex_lock (query->queue_lock);
        
        switch (type) {
        case KOZO_QUERY_NOTIFY_ADDED:
                queue = &query->added_queue;
                func = query->entries_added;
                break;
        case KOZO_QUERY_NOTIFY_CHANGED:
                queue = &query->changed_queue;
                func = query->entries_changed;
                break;
        case KOZO_QUERY_NOTIFY_REMOVED:
                queue = &query->removed_queue;
                func = query->entries_removed;
                break;        
        default:
                g_assert_not_reached ();
                return;
        }

        old_queue = *queue;
        *queue = kozo_entry_array_new (INITIAL_QUEUE_SIZE);

        g_mutex_unlock (query->queue_lock);

        if (func) {
                func (query,
                      old_queue->array,
                      old_queue->len,
                      query->user_data);
        }

        kozo_entry_array_free (old_queue);
}

void
kozo_query_notify (KozoQuery          *query,
                   KozoQueryNotifyType type,
                   TDB_DATA            key,
                   TDB_DATA            data)
{
        guint8 *dest;
        guint32 length;
        KozoEntryArray *queue;

        g_mutex_lock (query->queue_lock);

        switch (type) {
        case KOZO_QUERY_NOTIFY_ADDED:
                queue = query->added_queue;
                break;
        case KOZO_QUERY_NOTIFY_CHANGED:
                queue = query->changed_queue;
                break;
        case KOZO_QUERY_NOTIFY_REMOVED:
                queue = query->removed_queue;
                break;
        default:
                g_assert_not_reached ();
                return;
        }

        if (!query->notify (query, queue, key, data, &dest, &length)) {
                g_mutex_unlock (query->queue_lock);
                return;
        }

        kozo_entry_array_append (queue, dest, length);

        g_mutex_unlock (query->queue_lock);

        if (queue->len == QUEUE_THRESHOLD)
                flush_queue (query, type);
}

void
kozo_query_flush (KozoQuery *query)
{
        flush_queue (query, KOZO_QUERY_NOTIFY_ADDED);
        flush_queue (query, KOZO_QUERY_NOTIFY_CHANGED);
        flush_queue (query, KOZO_QUERY_NOTIFY_REMOVED);
}
