/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <errno.h>
#include <unistd.h>
#include <sys/syscall.h>

#include <glib.h>
#include <gio/gio.h>

#include <dbus/dbus-protocol.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include "bkl-investigator.h"
#include "bkl-investigator-dbus.h"

#include "metadata-defines.h"

typedef struct _InvestigateUri {
    char *source;
    char *uri;
} InvestigateUri;

struct _BklInvestigator {
    DBusGConnection *connection;
    BklInvestigatorDBus *dbus;

    GQueue *pending;
    guint32 worker_id;
    guint32 death_id;

    GMainLoop *mainloop;
};

static BklInvestigator *investigator = NULL;

#undef HAVE_IO_PRIO
#if defined (__i386__)
#  define HAVE_IO_PRIO
#  define __NR_ioprio_set 289
#elif defined (__x86_64__)
#  define HAVE_IO_PRIO
#  define __NR_ioprio_set 251
#else
#  warn "Architecture does not support ioprio modification"
#endif
#define IOPRIO_WHO_PROCESS 1
#define IOPRIO_CLASS_IDLE 3
#define IOPRIO_CLASS_SHIFT 13
#define IOPRIO_IDLE_LOWEST (7 | (IOPRIO_CLASS_IDLE << IOPRIO_CLASS_SHIFT))

#define INVESTIGATOR_NAME "org.moblin.Bickley.Investigator"
#define INVESTIGATOR_PATH "/org/moblin/Bickley/Investigator"
#define INVESTIGATOR_IFACE "org.moblin.Bickley.Investigator"

#define DEATH_TIMEOUT 10 /* seconds from the last file is processed to death */

/* FIXME: Whats the difference between time::modified and time::changed? */
#define REQUIRED_FIELDS "standard::size,standard::content-type,time::modified"

static gboolean
get_audio_metadata (GFile      *file,
                    GFileInfo  *info,
                    const char *mimetype,
                    GHashTable *metadata)
{
    /* If quick metadata fails (it only handles common types), then we try
       to use GStreamer */
    if (bkl_task_quick_audio_get_metadata (file, info,
                                           mimetype, metadata) == FALSE) {
        /*
          bkl_task_gstreamer_audio_get_metadata (file, info,
          mimetype, metadata);
        */
        return FALSE;
    }

    return TRUE;
}

static gboolean
get_image_metadata (GFile      *file,
                    GFileInfo  *info,
                    const char *mimetype,
                    GHashTable *metadata)
{
    gboolean care = FALSE;

    care |= bkl_task_pixbuf_get_metadata (file, info, mimetype, metadata);
    if (care) {
        bkl_task_pixbuf_thumbnail (file, info, mimetype, metadata);
    }
    care |= bkl_task_exif_get_metadata (file, info, mimetype, metadata);

    return care;
}

static gboolean
get_video_metadata (GFile      *file,
                    GFileInfo  *info,
                    const char *mimetype,
                    GHashTable *metadata)
{
    /*bkl_task_video_get_metadata (file, info, mimetype, metadata);*/
    return FALSE;
}

static GHashTable *
process_uri (const char *uri)
{
    GFile *file;
    GFileInfo *info;
    GHashTable *metadata;
    GError *error = NULL;
    const char *mimetype;
    GTimeVal result;
    gboolean care = FALSE;

    metadata = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
    file = g_file_new_for_uri (uri);
    info = g_file_query_info (file, REQUIRED_FIELDS, G_FILE_QUERY_INFO_NONE,
                              NULL, &error);
    if (info == NULL) {
        g_warning ("Error getting %s info: %s", uri, error->message);
        g_error_free (error);

        return metadata;
    }

    mimetype = g_file_info_get_content_type (info);

    /* Go through the various tasks */
    if (get_audio_metadata (file, info, mimetype, metadata)) {
        g_hash_table_insert (metadata, METADATA_TYPE, g_strdup (AUDIO_TYPE));
        care = TRUE;
    }
    if (get_image_metadata (file, info, mimetype, metadata)) {
        g_hash_table_insert (metadata, METADATA_TYPE, g_strdup (IMAGE_TYPE));
        care = TRUE;
    }
    if (get_video_metadata (file, info, mimetype, metadata)) {
        g_hash_table_insert (metadata, METADATA_TYPE, g_strdup (VIDEO_TYPE));
        care = TRUE;
    }

    if (care == FALSE) {
        g_object_unref (info);
        g_object_unref (file);
        return metadata;
    }

    g_hash_table_insert (metadata, METADATA_FILE_URI, g_strdup (uri));

    g_hash_table_insert (metadata, METADATA_FILE_SIZE,
                         g_strdup_printf ("%lld", g_file_info_get_size (info)));

    g_file_info_get_modification_time (info, &result);
    g_hash_table_insert (metadata, METADATA_FILE_MTIME,
                         g_strdup_printf ("%ld", result.tv_sec));

    g_print ("new mtime: %ld\n", result.tv_sec);
    g_hash_table_insert (metadata, METADATA_FILE_MIMETYPE, g_strdup (mimetype));


    g_object_unref (info);
    g_object_unref (file);

    return metadata;
}

static gboolean
kill_investigator (gpointer userdata)
{
    g_print ("Killing...\n");
    g_main_loop_quit (investigator->mainloop);
    return FALSE;
}

static void
dump_metadata (gpointer key,
               gpointer value,
               gpointer userdata)
{
    g_print ("%s: %s\n", key, value);
}

static gboolean
worker (gpointer userdata)
{
    InvestigateUri *iu;
    char *uri;
    GHashTable *metadata;

    iu = g_queue_pop_head (investigator->pending);
    if (G_UNLIKELY (iu == NULL)) {
        /* This shouldn't happen, but just in case */
        investigator->worker_id = 0;
        return FALSE;
    }

    /* Let the orbiter know we're doing something */
    bkl_investigator_dbus_processing (investigator->dbus, iu->source, iu->uri);

    metadata = process_uri (iu->uri);

    /* g_hash_table_foreach (metadata, dump_metadata, NULL); */

    bkl_investigator_dbus_completed (investigator->dbus, iu->source,
                                     iu->uri, metadata);

    g_hash_table_destroy (metadata);
    g_free (iu->uri);
    g_free (iu->source);
    g_free (iu);

    if (g_queue_peek_head (investigator->pending)) {
        return TRUE;
    } else {
        /* When we've run out of things to investigate, we stop the worker
           and start a timeout til shutdown. If new uris are added before
           DEATH_TIMEOUT expires, it is stopped and the worker is started
           again */
        g_print ("Death approaches\n");
        investigator->death_id = g_timeout_add_seconds (DEATH_TIMEOUT,
                                                        kill_investigator,
                                                        NULL);
        investigator->worker_id = 0;
        return FALSE;
    }
}

static gboolean
dbus_init (void)
{
    DBusGProxy *proxy;
    guint32 request_name_ret;
    GError *error = NULL;

    investigator->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!investigator->connection) {
        return FALSE;
    }

    proxy = dbus_g_proxy_new_for_name (investigator->connection,
                                       DBUS_SERVICE_DBUS,
                                       DBUS_PATH_DBUS,
                                       DBUS_INTERFACE_DBUS);
    if (!org_freedesktop_DBus_request_name (proxy, INVESTIGATOR_NAME,
                                            0, &request_name_ret, &error)) {
        return FALSE;
    }

    if (request_name_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
        return FALSE;
    }

    investigator->dbus = bkl_investigator_dbus_new ();
    dbus_g_connection_register_g_object (investigator->connection,
                                         INVESTIGATOR_PATH,
                                         (GObject *) investigator->dbus);

    return TRUE;
}

static void
ensure_thumbnail_dir (void)
{
    char *thumbdir;

    thumbdir = g_build_filename (g_get_home_dir (), ".bkl-thumbnails", NULL);
    g_mkdir_with_parents (thumbdir, 0777);
}

int
main (int    argc,
      char **argv)
{
    gst_init (&argc, &argv);

    ensure_thumbnail_dir ();

    /* We're a background service, make sure we run on a nice level */
    nice (19);
#ifdef HAVE_IO_PRIO
    if (syscall(__NR_ioprio_set,
                IOPRIO_WHO_PROCESS, 0, IOPRIO_IDLE_LOWEST) == -1) {
        g_warning ("Cannot set IO priority to idle class: %s (%d)",
                   g_strerror (errno), errno);
    }
#endif

    investigator = g_new0 (BklInvestigator, 1);
    if (dbus_init () == FALSE) {
        /* Investigator is already running...quit */
        return 0;
    }

    investigator->pending = g_queue_new ();

    investigator->mainloop = g_main_loop_new (NULL, TRUE);
    g_main_loop_run (investigator->mainloop);

    return 0;
}

void
bkl_investigator_add_uris (const char  *source,
                           const char **uris,
                           int          n_uris)
{
    int i;

    if (n_uris == 0) {
        return;
    }

    /* Stop the killer */
    if (investigator->death_id) {
        g_print ("Stopping death\n");
        g_source_remove (investigator->death_id);
        investigator->death_id = 0;
    }

    for (i = 0; uris[i]; i++) {
        InvestigateUri *iu = g_new (InvestigateUri, 1);
        iu->source = g_strdup (source);
        iu->uri = g_strdup (uris[i]);

        g_print ("(%s) - Adding %s\n", iu->source, iu->uri);
        g_queue_push_tail (investigator->pending, iu);
    }

    /* Start the worker if its not running */
    if (investigator->worker_id == 0) {
        investigator->worker_id = g_idle_add (worker, NULL);
    }
}
