/*
 * Bickley - a meta data management framework.
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>
#include <errno.h>
#include <signal.h>

#include <glib.h>
#include <kozo.h>
#include <gconf/gconf-client.h>

#include <libgupnp/gupnp-context.h>
#include <libgupnp/gupnp-control-point.h>
#include <libgupnp/gupnp-device-info.h>

#include <dbus/dbus-protocol.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <bickley/bkl-db.h>
#include <bickley/bkl-entry.h>
#include <bickley/bkl-item-audio.h>
#include <bickley/bkl-item-image.h>
#include <bickley/bkl-item-video.h>

#include "bkl-orbiter.h"
#include "bkl-orbiter-dbus.h"
#include "bkl-path-finder.h"
#include "bkl-investigator-bindings.h"
#include "bkl-marshal.h"
#include "bkl-source-gconf.h"
#include "bkl-source-removable.h"
#include "bkl-source-upnp.h"

#include "metadata-defines.h"

struct _BklOrbiter {
    BklOrbiterDBus *dbus;
    DBusGConnection *connection;
    DBusGProxy *proxy;

    GConfClient *gconf;
    GMainLoop *mainloop;

    KozoDB *db;

    GVolumeMonitor *volume_monitor;

    BklSource *local_source;
    GHashTable *sources;

    guint32 worker_id;

    /* These are the uris that we've told the investigator to look at
       and it hasn't got to yet */
    GHashTable *requested_uris;
    char *processing_uri;

    DBusGProxy *investigator;
    GPid investigator_pid;
    guint32 investigator_processing_id;

    GUPnPContext *upnp_monitor;
    GUPnPControlPoint *upnp_cp;
};

#define ORBITER_NAME "org.moblin.Bickley.Orbiter"
#define ORBITER_PATH "/org/moblin/Bickley/Orbiter"
#define ORBITER_IFACE "org.moblin.Bickley.Orbiter"

#define INVESTIGATOR_NAME "org.moblin.Bickley.Investigator"
#define INVESTIGATOR_PATH "/org/moblin/Bickley/Investigator"
#define INVESTIGATOR_IFACE "org.moblin.Bickley.Investigator"

#define PROCESSING_ALLOWED_TIMEOUT 10 /* Seconds */

static BklOrbiter *orbiter = NULL;
static gboolean running_investigator = FALSE;

static void
kill_investigator (void)
{
    if (running_investigator && orbiter->investigator_pid) {
        if (kill (orbiter->investigator_pid, SIGKILL) == -1) {
            g_warning ("Error killing the investigator (%d): %s (%d)",
                       orbiter->investigator_pid, strerror (errno), errno);
            orbiter->investigator_pid = 0;
            running_investigator = FALSE;
        }
    }
}

static gboolean
processing_expired (gpointer userdata)
{
    g_warning ("Investigator has not processed anything for 10 seconds...killing");
    kill_investigator ();

    return FALSE;
}

static void
investigator_processing (DBusGProxy *proxy,
                         const char *source_name,
                         const char *current_uri,
                         gpointer    userdata)
{
    if (orbiter->investigator_processing_id > 0) {
        g_source_remove (orbiter->investigator_processing_id);
    }

    /* We take this out of the hashtable so that if it crashes/hangs
       the investigator, it won't be looked at again when we restart

       FIXME: Should we allow a uri multiple tries before failing it? */
    g_hash_table_remove (orbiter->requested_uris, current_uri);

    orbiter->investigator_processing_id = g_timeout_add_seconds
        (PROCESSING_ALLOWED_TIMEOUT, processing_expired, NULL);
}

static void
dump_metadata (gpointer key,
               gpointer value,
               gpointer userdata)
{
    g_print ("%s: %s\n", key, value);
}

static BklItem *
make_audio_item (const char *uri,
                 GHashTable *metadata)
{
    BklItemAudio *item;
    const char *data;
    int number;

    item = bkl_item_audio_new ();

    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_TITLE)) {
        bkl_item_audio_set_title (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_ALBUM)) {
        bkl_item_audio_set_album (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_ARTIST)) {
        GPtrArray *array = g_ptr_array_sized_new (1);

        g_ptr_array_add (array, g_strdup (data));
        bkl_item_audio_set_artists (item, array);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_GENRE)) {
        bkl_item_audio_set_genre (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_COMMENT)) {
        bkl_item_audio_set_comment (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_YEAR)) {
        bkl_item_audio_set_year (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_DURATION)) {
        number = atoi (data);
        bkl_item_audio_set_duration (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_TRACK)) {
        number = atoi (data);
        bkl_item_audio_set_track (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_TOTAL_TRACKS)) {
        number = atoi (data);
        bkl_item_audio_set_maxtracks (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_AUDIO_DISC)) {
        number = atoi (data);
        bkl_item_audio_set_disc (item, number);
    }

    return (BklItem *) item;
}

static BklItem *
make_image_item (const char *uri,
                 GHashTable *metadata)
{
    BklItemImage *item;
    const char *data;
    int number;

    item = bkl_item_image_new ();

    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_TITLE)) {
        bkl_item_image_set_title (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_MODEL)) {
        bkl_item_image_set_camera_model (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_MAKE)) {
        bkl_item_image_set_camera_make (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_DATE_TIME)) {
        bkl_item_image_set_time (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_DATE_TIME_ORIGINAL)) {
        bkl_item_image_set_time_original (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_DATE_TIME_DIGITIZED)) {
        bkl_item_image_set_time_digitized (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_ORIENTATION)) {
        bkl_item_image_set_orientation (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_EXPOSURE_TIME)) {
        bkl_item_image_set_exposure (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_APERTURE_VALUE)) {
        bkl_item_image_set_aperture (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_METERING_MODE)) {
        bkl_item_image_set_metering_mode (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_FLASH)) {
        bkl_item_image_set_flash (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_FOCAL_LENGTH)) {
        bkl_item_image_set_focal_length (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_SHUTTER_SPEED)) {
        bkl_item_image_set_shutter_speed (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_ISO)) {
        bkl_item_image_set_iso (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_SOFTWARE)) {
        bkl_item_image_set_software (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_COMMENT)) {
        bkl_item_image_set_comment (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_USER_COMMENT)) {
        bkl_item_image_set_user_comment (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_AUTHOR)) {
        bkl_item_image_set_author (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_COPYRIGHT)) {
        bkl_item_image_set_copyright (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_KEYWORDS)) {
        bkl_item_image_set_keywords (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_SUBJECT)) {
        bkl_item_image_set_subject (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_WIDTH)) {
        number = atoi (data);
        bkl_item_image_set_width (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_IMAGE_HEIGHT)) {
        number = atoi (data);
        bkl_item_image_set_height (item, number);
    }

    return (BklItem *) item;
}

static BklItem *
make_video_item (const char *uri,
                 GHashTable *metadata)
{
    BklItemVideo *item;
    const char *data;
    int number;

    item = bkl_item_video_new ();

    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_TITLE)) {
        bkl_item_video_set_title (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_DIRECTOR)) {
        bkl_item_video_set_directory (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_SERIES_NAME)) {
        bkl_item_video_set_series_name (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_SEASON)) {
        number = atoi (data);
        bkl_item_video_set_season (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_EPISODE)) {
        number = atoi (data);
        bkl_item_video_set_episode (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_YEAR)) {
        number = atoi (data);
        bkl_item_video_set_year (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_WIDTH)) {
        number = atoi (data);
        bkl_item_video_set_width (item, number);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_VIDEO_HEIGHT)) {
        number = atoi (data);
        bkl_item_video_set_height (item, number);
    }

    return (BklItem *) item;
}

static void
add_file_metadata (BklItem    *item,
                   GHashTable *metadata)
{
    const char *data;

    if (data = g_hash_table_lookup (metadata, METADATA_FILE_URI)) {
        bkl_item_set_uri (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_FILE_MIMETYPE)) {
        bkl_item_set_mimetype (item, data);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_FILE_MTIME)) {
        glong mtime = strtol (data, NULL, 10);
        bkl_item_set_modification_time (item, mtime);
    }
    if (data = g_hash_table_lookup (metadata, METADATA_FILE_SIZE)) {
        goffset size = strtoull (data, NULL, 10);
        bkl_item_set_size (item, size);
    }
}

static void
investigator_completed (DBusGProxy *proxy,
                        const char *source_name,
                        const char *uri,
                        GHashTable *metadata,
                        gpointer    userdata)
{
    BklSource *source;
    BklItem *item = NULL, *old = NULL;
    const char *type;
    GError *error = NULL;

    /* Turn off any processing request */
    if (orbiter->investigator_processing_id > 0) {
        g_source_remove (orbiter->investigator_processing_id);
        orbiter->investigator_processing_id = 0;
    }

    if (g_hash_table_size (metadata) == 0) {
        /* This wasn't a filetype we really cared about */
        return;
    }

    type = g_hash_table_lookup (metadata, METADATA_TYPE);
    if (type == NULL) {
        return;
    }

    if (g_str_equal (type, AUDIO_TYPE)) {
        item = make_audio_item (uri, metadata);
    } else if (g_str_equal (type, IMAGE_TYPE)) {
        item = make_image_item (uri, metadata);
    } else if (g_str_equal (type, VIDEO_TYPE)) {
        item = make_image_item (uri, metadata);
    } else {
        g_warning ("Unknown type: %s", type);
        return;
    }

    add_file_metadata (item, metadata);
    /* g_hash_table_foreach (metadata, dump_metadata, NULL); */

    source = g_hash_table_lookup (orbiter->sources, source_name);
    if (source == NULL) {
        g_warning ("Error getting source: %s", source_name);
        return;
    }

    bkl_source_add_item (source, uri, item, &error);
    g_object_unref (item);
}

static gboolean
ensure_investigator (void)
{
    if (!running_investigator) {
        GError *error = NULL;
        guint start_ret;

        if (!org_freedesktop_DBus_start_service_by_name (orbiter->proxy,
                                                         INVESTIGATOR_NAME, 0,
                                                         &start_ret, &error)) {
            g_warning ("Investigator has failed to start: %s",
                       error->message);
            g_error_free (error);

            running_investigator = FALSE;
            return FALSE;
        }

        running_investigator = TRUE;

        orbiter->investigator = dbus_g_proxy_new_for_name (orbiter->connection,
                                                           INVESTIGATOR_NAME,
                                                           INVESTIGATOR_PATH,
                                                           INVESTIGATOR_IFACE);
        dbus_g_proxy_add_signal (orbiter->investigator, "Processing",
                                 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
        dbus_g_proxy_connect_signal (orbiter->investigator, "Processing",
                                     G_CALLBACK (investigator_processing),
                                     NULL, NULL);
        dbus_g_proxy_add_signal (orbiter->investigator, "UriCompleted",
                                 G_TYPE_STRING, G_TYPE_STRING,
                                 DBUS_TYPE_G_STRING_STRING_HASHTABLE,
                                 G_TYPE_INVALID);
        dbus_g_proxy_connect_signal (orbiter->investigator, "UriCompleted",
                                     G_CALLBACK (investigator_completed),
                                     NULL, NULL);

        if (!org_moblin_Bickley_Investigator_get_process_id
            (orbiter->investigator, &orbiter->investigator_pid, &error)) {
            g_warning ("Error getting investigator PID: %s", error->message);
            g_error_free (error);
            orbiter->investigator_pid = 0;
            return FALSE;
        }
    }

    return TRUE;
}

static void
name_owner_changed (DBusGProxy *proxy,
                    const char *name,
                    const char *old_owner,
                    const char *new_owner,
                    gpointer    userdata)
{
    int size;

    if (strcmp (name, INVESTIGATOR_NAME) != 0) {
        /* Only need to listen for investigator changing */
        return;
    }

    if (new_owner && *new_owner != 0) {
        /* Investigator has just been owned so everything is good */
        return;
    }

    /* New owner is empty, so the investigator has left the bus
       for some reason. Check to see if it needs restarted */
    if (orbiter->investigator_processing_id) {
        g_source_remove (orbiter->investigator_processing_id);
        orbiter->investigator_processing_id = 0;
    }

    orbiter->investigator_pid = 0;
    g_object_unref (orbiter->investigator);
    orbiter->investigator = NULL;
    running_investigator = FALSE;

    size = g_hash_table_size (orbiter->requested_uris);
    if (size > 0) {
        /* We were still waiting to hear back from the investigator about
           some uris, so it has not finished its work. Need to remove the
           last uri it was trying and restart it */
        ensure_investigator ();

        /* FIXME: Feed the remaining uris to the investigator again */
    }
}

void
bkl_orbiter_investigate_files (BklSource *source,
                               GPtrArray *files)
{
    GPtrArray *fa;
    GError *error = NULL;
    int i;

    fa = g_ptr_array_new ();
    i = 0;
    while (i < files->len) {
        char *uri = files->pdata[i];
        gpointer value;

        /* If we're still waiting for the uri to be indexed, ignore it */
        if (g_hash_table_lookup (orbiter->requested_uris, uri) != NULL) {
            g_free (uri);
            i++;

            continue;
        }

        g_print ("Investigating %s\n", uri);
        g_ptr_array_add (fa, uri);
        g_hash_table_insert (orbiter->requested_uris, uri, NULL);
        i++;
    }

    if (fa->len == 0) {
        g_ptr_array_free (fa, TRUE);
        return;
    }

    /* NULL terminate the array */
    g_ptr_array_add (fa, NULL);

    if (running_investigator == FALSE) {
        if (ensure_investigator () == FALSE) {
            for (i = 0; i < fa->len; i++) {
                g_free (fa->pdata[i]);
            }
            g_ptr_array_free (fa, TRUE);

            return;
        }
    }

    if (!org_moblin_Bickley_Investigator_investigate_uris
        (orbiter->investigator, source->name, (const char **) fa->pdata,
         fa->len, &error)) {
        g_warning ("Error telling investigator about %d uris: %s", fa->len,
                   error->message);
        g_error_free (error);
    }

    /* Don't need to free the uris in @fa as they should be freed when
       removed from the requested_uri hashtable */
    g_ptr_array_free (fa, TRUE);
}

void
bkl_orbiter_remove_files (GPtrArray *files)
{
    int i;

    for (i = 0; i < files->len; i++) {
        GError *error = NULL;
        char *uri;

        uri = g_file_get_uri (files->pdata[i]);
        g_print ("Deleted %s\n", uri);
        bkl_db_remove (orbiter->db, uri, &error);
        if (error) {
            g_warning ("Error removing %s: %s", uri, error->message);
            g_error_free (error);
        }
        g_free (uri);
    }
}

static gboolean
dbus_init (void)
{
    guint32 request_name_ret;
    GError *error = NULL;

    /* Register the weird marshaller needed */
    dbus_g_object_register_marshaller (bkl_marshal_VOID__STRING_STRING,
                                       G_TYPE_NONE, G_TYPE_STRING,
                                       G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_object_register_marshaller (bkl_marshal_VOID__STRING_STRING_BOXED,
                                       G_TYPE_NONE, G_TYPE_STRING,
                                       G_TYPE_STRING, G_TYPE_BOXED,
                                       G_TYPE_INVALID);

    orbiter->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (orbiter->connection == NULL) {
        g_warning ("Error getting bus: %s", error->message);
        g_error_free (error);

        return FALSE;
    }

    orbiter->proxy = dbus_g_proxy_new_for_name (orbiter->connection,
                                                DBUS_SERVICE_DBUS,
                                                DBUS_PATH_DBUS,
                                                DBUS_INTERFACE_DBUS);
    if (!org_freedesktop_DBus_request_name (orbiter->proxy, ORBITER_NAME,
                                            0, &request_name_ret, &error)) {
        g_warning ("Error registering DBus: %s", error->message);
        g_error_free (error);

        return FALSE;
    }

    if (request_name_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
        return FALSE;
    }

    dbus_g_proxy_add_signal (orbiter->proxy, "NameOwnerChanged",
                             G_TYPE_STRING, G_TYPE_STRING,
                             G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal (orbiter->proxy, "NameOwnerChanged",
                                 G_CALLBACK (name_owner_changed), NULL, NULL);

    orbiter->dbus = bkl_orbiter_dbus_new ();
    dbus_g_connection_register_g_object (orbiter->connection, ORBITER_PATH,
                                         (GObject *) orbiter->dbus);
    return TRUE;
}

static void
kozo_shutdown (void)
{
    /* We want to close the DB cleanly */
    if (orbiter->db) {
        kozo_db_unref (orbiter->db);
    }

    kozo_index_shutdown ();
    kozo_db_shutdown ();
    kozo_expr_shutdown ();
    kozo_field_shutdown ();
    kozo_entry_shutdown ();
    kozo_util_shutdown ();
}

static gboolean
kozo_init (void)
{
    GError *error = NULL;

    kozo_util_init ();
    kozo_entry_init ();
    kozo_field_init ();
    kozo_expr_init ();
    kozo_db_init ();
    kozo_index_init ();

    return TRUE;
}

static void
get_more_work (gpointer key,
               gpointer value,
               gpointer userdata)
{
    BklSource *source = (BklSource *) value;
    gboolean *more_work = userdata;

    g_print ("(%s): working\n", source->name);
    *more_work |= bkl_source_do_work (source);
}

static gboolean
worker (gpointer userdata)
{
    gboolean more_work = FALSE;

    g_hash_table_foreach (orbiter->sources, get_more_work, &more_work);

    if (more_work == FALSE) {
        /* Turn off the orbitor */
        orbiter->worker_id = 0;
        g_print ("Turning off worker\n");
        return FALSE;
    } else {
        /* Sleep for a bit before processing queue again */
        /* g_usleep (G_USEC_PER_SEC * 0.03); */

        return TRUE;
    }
}

void
bkl_orbiter_start_worker (void)
{
    if (orbiter->worker_id > 0) {
        return;
    }

    orbiter->worker_id = g_idle_add (worker, NULL);
}

static void
mount_added (GVolumeMonitor *volume_monitor,
             GMount         *mount,
             gpointer        userdata)
{
    BklSource *source;
    GFile *root;
    const char *filename;

    root = g_mount_get_root (mount);
    source = bkl_source_removable_new (mount);
    g_hash_table_insert (orbiter->sources, g_file_get_path (root), source);
    g_object_unref (root);

    filename = kozo_db_get_filename (source->db);
    bkl_orbiter_dbus_source_added (orbiter->dbus, source->name, filename);
}

/* We want to listen for the pre-unmounted signal, rather than the removed
   signal as we need to destroy the source to allow the mount to be removed */
static void
mount_pre_unmount (GVolumeMonitor *volume_monitor,
                   GMount         *mount,
                   gpointer        userdata)
{
    BklSource *source;
    const char *filename;
    GFile *root;
    char *path;

    g_print ("Pre-unmounted\n");
    root = g_mount_get_root (mount);
    path = g_file_get_path (root);

    source = g_hash_table_lookup (orbiter->sources, path);
    g_object_unref (root);

    if (source == NULL) {
        g_free (path);
        return;
    }

    g_hash_table_remove (orbiter->sources, path);
    g_free (path);

    filename = kozo_db_get_filename (source->db);
    bkl_orbiter_dbus_source_removed (orbiter->dbus, source->name, filename);

    bkl_source_destroy (source);
}

static void
init_volume_monitor (void)
{
    GList *mounts, *m;

    orbiter->volume_monitor = g_volume_monitor_get ();
    g_signal_connect (orbiter->volume_monitor, "mount-added",
                      G_CALLBACK (mount_added), NULL);
    g_signal_connect (orbiter->volume_monitor, "mount-pre-unmount",
                      G_CALLBACK (mount_pre_unmount), NULL);
#if 0
    g_signal_connect (orbiter->volume_monitor, "mount-removed",
                      G_CALLBACK (mount_removed), NULL);
#endif

    mounts = g_volume_monitor_get_mounts (orbiter->volume_monitor);
    for (m = mounts; m; m = m->next) {
        GMount *mount = m->data;

        mount_added (orbiter->volume_monitor, mount, NULL);
        g_object_unref (mount);
    }
    g_list_free (mounts);
}

static void
upnp_device_available (GUPnPControlPoint *cp,
                       GUPnPDeviceProxy  *proxy,
                       gpointer           userdata)
{
    GUPnPDeviceInfo *info = (GUPnPDeviceInfo *) proxy;
    BklSource *source;
    const char *udn, *filename;

    source = bkl_source_upnp_new (proxy);

    udn = gupnp_device_info_get_udn (info);
    g_hash_table_insert (orbiter->sources, g_strdup (udn), source);

    filename = kozo_db_get_filename (source->db);
    bkl_orbiter_dbus_source_added (orbiter->dbus, source->name, filename);
}

static void
upnp_device_unavailable (GUPnPControlPoint *cp,
                         GUPnPDeviceProxy  *proxy,
                         gpointer           userdata)
{
    GUPnPDeviceInfo *info = (GUPnPDeviceInfo *) proxy;
    BklSource *source;
    const char *udn, *filename;

    udn = gupnp_device_info_get_udn (info);
    source = g_hash_table_lookup (orbiter->sources, udn);

    if (source == NULL) {
        return;
    }

    g_hash_table_remove (orbiter->sources, udn);

    filename = kozo_db_get_filename (source->db);
    bkl_orbiter_dbus_source_removed (orbiter->dbus, source->name, filename);

    bkl_source_destroy (source);
}

static void
init_upnp_monitor (void)
{
#if 0
    orbiter->upnp_monitor = gupnp_context_new (NULL, NULL, 0, NULL);
    orbiter->upnp_cp = gupnp_control_point_new
        (orbiter->upnp_monitor, "urn:schemas-upnp-org:device:MediaServer:1");
    g_signal_connect (orbiter->upnp_cp, "device-proxy-available",
                      G_CALLBACK (upnp_device_available), NULL);
    g_signal_connect (orbiter->upnp_cp, "device-proxy-unavailable",
                      G_CALLBACK (upnp_device_unavailable), NULL);

    /* Start UPnP searching */
    gssdp_resource_browser_set_active
        (GSSDP_RESOURCE_BROWSER (orbiter->upnp_cp), TRUE);
#endif
}

int
main (int    argc,
      char **argv)
{
    g_thread_init (NULL);
    g_type_init ();

    orbiter = g_new0 (BklOrbiter, 1);
    if (dbus_init () == FALSE) {
        /* Orbiter is already running, so just quit */
        return 0;
    }

    if (kozo_init () == FALSE) {
        /* Error initializing Kozo */
        return 0;
    }

    orbiter->requested_uris = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                     g_free, NULL);
    orbiter->sources = g_hash_table_new (g_str_hash, g_str_equal);

    orbiter->local_source = bkl_source_gconf_new ("local-media");
    g_hash_table_insert (orbiter->sources, orbiter->local_source->name,
                         orbiter->local_source);

    init_volume_monitor ();
    init_upnp_monitor ();

    /* This will run through the sources to see if there is work to do
       and if not, it will turn itself off */
    bkl_orbiter_start_worker ();

    orbiter->mainloop = g_main_loop_new (NULL, TRUE);
    g_main_loop_run (orbiter->mainloop);

    kozo_shutdown ();

    return 0;
}

GList *
bkl_orbiter_get_sources (void)
{
    GList *sources;

    sources = g_hash_table_get_values (orbiter->sources);
    return sources;
}
