/*
 * Bickley - a meta data management framework, the file system spider.
 *
 * Copyright © 2007-2008 Øyvind Kolås <pippin@gimp.org>
 * Copyright © 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <glib.h>
#include <gio/gio.h>

#include "bkl-orbiter.h"
#include "bkl-path-finder.h"

/*#define REQUIRED_FIELDS "*" */
#define REQUIRED_FIELDS "standard::type,standard::name,time::modified"

static void
unref_pending (gpointer data,
               gpointer userdata)
{
    g_object_unref ((GObject *) data);
}

void
bkl_path_finder_destroy (BklFinder *finder)
{
    BklPathFinder *self = (BklPathFinder *) finder;

    g_free (finder->base_uri);
    g_object_unref (self->base_file);

    g_queue_foreach (self->pending, unref_pending, NULL);
    g_queue_free (self->pending);

    g_hash_table_destroy (self->monitors);
    g_free (self);
}

static void
directory_changed_cb (GFileMonitor     *monitor,
                      GFile            *file,
                      GFile            *other_file,
                      GFileMonitorEvent event,
                      BklPathFinder    *finder)
{
    GFileInfo *info;
    GError *error = NULL;
    InvestigateUri *iu;
    GPtrArray *fa;
    char *uri;

    switch (event) {
    case G_FILE_MONITOR_EVENT_CHANGED:
    case G_FILE_MONITOR_EVENT_CREATED:
        break;

    case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
        info = g_file_query_info (file, "standard::type",
                                  G_FILE_QUERY_INFO_NONE, NULL, &error);
        if (error != NULL) {
            uri = g_file_get_uri (file);

            g_warning ("Error getting info for %s: %s", uri, error->message);
            g_error_free (error);
        }

        if (info && g_file_info_get_file_type (info) == G_FILE_TYPE_DIRECTORY) {
            /* We ignore directories as all the changes will be dealt with in
               other events */
            return;
        }
        g_queue_push_tail (finder->pending, g_object_ref (file));

        /* Tell the orbiter that we have more work to do */
        bkl_orbiter_start_worker ();
        break;

        break;

    case G_FILE_MONITOR_EVENT_DELETED:
        fa = g_ptr_array_sized_new (1);
        g_ptr_array_add (fa, file);

        bkl_orbiter_remove_files (fa);
        g_ptr_array_free (fa, TRUE);

        /* We remove any monitor that might be active on this file */
        uri = g_file_get_uri (file);
        g_hash_table_remove (finder->monitors, uri);
        g_free (uri);
        break;

    case G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED:
        break;

    case G_FILE_MONITOR_EVENT_PRE_UNMOUNT:
    case G_FILE_MONITOR_EVENT_UNMOUNTED:
    default:
        break;
    }
}

static void
process_file (BklPathFinder *finder,
              GFile         *file)
{
    GError *error = NULL;
    GFileInfo *info;
    GFileType type;
    GPtrArray *flist = NULL;
    GTimeVal mtime;
    int i;

    info = g_file_query_info (file, REQUIRED_FIELDS,
                              G_FILE_QUERY_INFO_NONE, NULL, &error);
    if (info == NULL) {
        char *uri;

        uri = g_file_get_uri (file);
        g_warning ("Error querying info for %s: %s", uri, error->message);
        g_free (uri);

        g_error_free (error);

        return;
    }

    type = g_file_info_get_file_type (info);

    if (type == G_FILE_TYPE_REGULAR) {
        InvestigateUri *iu = g_new (InvestigateUri, 1);

        g_print ("Processing %s\n", g_file_get_uri (file));
        flist = g_ptr_array_sized_new (1);

        iu->file = g_object_ref (file);
        g_file_info_get_modification_time (info, &iu->mtime);
        g_ptr_array_add (flist, iu);

        g_object_unref (info);
    } else if (type == G_FILE_TYPE_DIRECTORY) {
        GFileEnumerator *fe;
        GFileMonitor *dir_monitor;
        char *uri;

        g_print ("Processing dir %s\n", g_file_get_uri (file));
        g_object_unref (info);

        uri = g_file_get_uri (file);
        fe = g_file_enumerate_children (file, REQUIRED_FIELDS,
                                        G_FILE_QUERY_INFO_NONE, NULL, &error);
        if (fe == NULL) {
            g_warning ("Error enumerating children of %s: %s", uri,
                       error->message);
            g_free (uri);

            g_error_free (error);

            return;
        }

        flist = g_ptr_array_new ();

        while (info = g_file_enumerator_next_file (fe, NULL, NULL)) {
            GFile *subfile;
            const char *filename;

            filename = g_file_info_get_name (info);

            /* Skip hidden files */
            if (filename && filename[0] == '.') {
                g_object_unref (info);
                continue;
            }

            subfile = g_file_get_child (file, filename);

            type = g_file_info_get_file_type (info);
            if (type == G_FILE_TYPE_REGULAR) {
                InvestigateUri *iu = g_new (InvestigateUri, 1);

                iu->file = subfile;
                g_file_info_get_modification_time (info, &iu->mtime);
                g_ptr_array_add (flist, iu);
                g_print ("Got file: %s\n", g_file_get_uri (subfile));
            } else if (type == G_FILE_TYPE_DIRECTORY) {
                g_queue_push_tail (finder->pending, subfile);
            }

            g_object_unref (info);
        }

        g_object_unref (fe);

        /* As this is a directory, we want to add a monitor to it
           if we haven't done so already */
        if (g_hash_table_lookup (finder->monitors, uri) == NULL) {
            dir_monitor = g_file_monitor_directory (file, G_FILE_MONITOR_NONE,
                                                    NULL, &error);
            if (dir_monitor == NULL) {
                g_warning ("Error adding monitor for %s: %s", uri,
                           error->message);
                g_error_free (error);
            } else {
                /* Store the monitor so we can delete it when the directory is
                   destroyed */
                g_hash_table_insert (finder->monitors, g_strdup (uri),
                                     dir_monitor);
                g_signal_connect (dir_monitor, "changed",
                                  G_CALLBACK (directory_changed_cb), finder);
            }
        }

        g_free (uri);
    }

    /* Tell the investigator to do something with the files */
    bkl_source_investigate_files (((BklFinder *) finder)->source, flist);

    /* Unref all the new files now that they've been sent to the investigator */
    for (i = 0; i < flist->len; i++) {
        InvestigateUri *iu = flist->pdata[i];

        g_object_unref (iu->file);
        g_free (iu);
    }
    g_ptr_array_free (flist, TRUE);
}

/* return TRUE if there is more work */
gboolean
bkl_path_finder_lazy_dig (BklFinder *finder)
{
    BklPathFinder *self = (BklPathFinder *) finder;
    GFile *next;

    /* Pluck an item from from the pending queue and process it
       If it is a file then we tell the inspector to look at it
       If it is a directory, then we add all the children directories
       to the pending queue, and tell the inspector about all the
       direct children files */
    next = g_queue_pop_head (self->pending);
    if (next == NULL) {
        return FALSE;
    }

    process_file (self, next);

    g_object_unref (next);
    return (g_queue_peek_head (self->pending) != NULL);
}

BklFinder *
bkl_path_finder_new (BklSource  *source,
                     const char *uri)
{
    BklPathFinder *path_finder;
    BklFinder *finder;

    path_finder = g_new0 (BklPathFinder, 1);
    finder = (BklFinder *) path_finder;

    finder->destroy = bkl_path_finder_destroy;
    finder->lazy_dig = bkl_path_finder_lazy_dig;

    finder->base_uri = g_strdup (uri);
    finder->source = source;

    path_finder->base_file = g_file_new_for_uri (uri);

    path_finder->pending = g_queue_new ();
    g_queue_push_tail (path_finder->pending,
                       g_object_ref (path_finder->base_file));

    path_finder->monitors = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                   g_free, g_object_unref);
    return finder;
}
