#include <errno.h>
#include <unistd.h>
#include <gio/gio.h>

#include <bickley/bkl-db.h>
#include <bickley/bkl-entry.h>
#include <bickley/bkl-item-extended.h>

#include "bkl-finder.h"
#include "bkl-orbiter.h"
#include "bkl-path-finder.h"
#include "bkl-source.h"
#include "bkl-source-removable.h"

struct _BklSourceRemovable {
    BklSource source;

    BklFinder *finder;

    /* List of the uris in the database and the
       time they were last indexed at */
    GHashTable *last_indexed;
};

static void
bkl_source_removable_destroy (BklSource *source)
{
    BklSourceRemovable *removable = (BklSourceRemovable *) source;

    kozo_db_unref (source->db);
    g_free (source->name);

    bkl_finder_destroy (removable->finder);
    g_hash_table_destroy (removable->last_indexed);

    g_free (removable);
}

static gboolean
bkl_source_removable_do_work (BklSource *source)
{
    BklSourceRemovable *removable = (BklSourceRemovable *) source;
    gboolean more_work;

    more_work = bkl_finder_lazy_dig (removable->finder);
    return more_work;
}

/* FIXME: Can this be shared in BklSource? */
static void
bkl_source_removable_investigate_uris (BklSource *source,
                                       GPtrArray *files)
{
    BklSourceRemovable *removable = (BklSourceRemovable *) source;
    GPtrArray *fa;
    int i;

    fa = g_ptr_array_new ();

    for (i = 0; i < files->len; i++) {
        InvestigateUri *iu = files->pdata[i];
        char *uri = g_file_get_uri (iu->file);
        gpointer value;

        value = g_hash_table_lookup (removable->last_indexed, uri);
        if (value != NULL) {
            if (iu->mtime.tv_sec <= GPOINTER_TO_INT (value)) {
                g_free (uri);
                i++;

                continue;
            }
        }

        g_ptr_array_add (fa, uri);
    }

    bkl_orbiter_investigate_files (source, fa);
    /* investigate_files takes ownership of the uris */
    g_ptr_array_free (fa, TRUE);
}

/* FIXME: Can this be shared in BklSource? */
static GPtrArray *
copy_string_array (GPtrArray *original)
{
    GPtrArray *clone;
    int i;

    if (original == NULL) {
        return NULL;
    }

    clone = g_ptr_array_sized_new (original->len);
    for (i = 0; i < original->len; i++) {
        g_ptr_array_add (clone, g_strdup (original->pdata[i]));
    }

    return clone;
}

static gboolean
bkl_source_removable_add_item (BklSource  *source,
                               const char *uri,
                               BklItem    *item,
                               GError    **error)
{
    BklSourceRemovable *removable = (BklSourceRemovable *) source;
    BklItem *old;

    old = bkl_db_get_item (source->db, uri, error);
    if (*error != NULL) {
        if ((*error)->code != KOZO_DB_ERROR_KEY_NOT_FOUND) {
            return FALSE;
        }

        g_error_free (*error);
        *error = NULL;
    }

    if (old == NULL) {
        bkl_db_add_item (source->db, uri, item, error);
    } else {
        BklItemExtended *ext = (BklItemExtended *) old;
        GPtrArray *tags;

                /* Copy the extended details from the existing item
           as these shouldn't change on update */
        bkl_item_extended_set_use_count
            ((BklItemExtended *) item, bkl_item_extended_get_use_count (ext));
        bkl_item_extended_set_last_used
            ((BklItemExtended *) item, bkl_item_extended_get_last_used (ext));
        bkl_item_extended_set_rating
            ((BklItemExtended *) item, bkl_item_extended_get_rating (ext));
        bkl_item_extended_set_pinned
            ((BklItemExtended *) item, bkl_item_extended_get_pinned (ext));

        tags = bkl_item_extended_get_tags (ext);
        bkl_item_extended_set_tags ((BklItemExtended *) item,
                                    copy_string_array (tags));

        bkl_db_replace_item (source->db, uri, item, error);
        g_object_unref (old);
    }

    return TRUE;
}

static void
get_last_indexed (KozoDB     *db,
                  const char *key,
                  KozoEntry  *entry,
                  gpointer    userdata)
{
    BklSourceRemovable *source = (BklSourceRemovable *) userdata;
    KozoField *field;
    const char *data;
    glong t;

    field = bkl_entry_get_field (entry, BKL_FIELD_FILE);
    if (field == NULL) {
        return;
    }

    data = bkl_file_field_get_string (field, BKL_FILE_FIELD_MODIFICATION_TIME);
    if (data == NULL || *data == '\0') {
        return;
    }

    t = strtol (data, NULL, 10);
    g_hash_table_insert (source->last_indexed, g_strdup (key),
                         GINT_TO_POINTER (t));
}

static gboolean
kozo_init (BklSourceRemovable *source,
           GMount             *mount)
{
    BklSource *s = (BklSource *) source;
    GError *error = NULL;
    char *dirname, *filename, *name, *enc, *path;
    GFile *root;

    root = g_mount_get_root (mount);
    name = g_mount_get_name (mount);
    enc = g_compute_checksum_for_string (G_CHECKSUM_MD5, name, -1);
    path = g_file_get_path (root);
    g_object_unref (root);

    if (g_access (path, R_OK | W_OK) == 0) {
        dirname = g_build_filename (path, ".kozo", "databases", NULL);

        if (g_mkdir_with_parents (dirname, 0755) < 0) {
            g_warning ("Error making dir %s", strerror (errno));
        }
        filename = g_build_filename (dirname, enc, NULL);
        g_free (enc);
        g_free (dirname);

        s->db = bkl_db_get_for_path (filename, name, &error);
    } else {
        g_print ("%s is not writable, falling back to local\n", path);
        s->db = bkl_db_get (name, &error);
    }

    g_free (path);
    g_free (filename);

    if (s->db == NULL) {
        g_warning ("Error getting Bickley database: %s", error->message);
        g_error_free (error);

        g_free (name);
        return FALSE;
    }

    s->name = name;

    source->last_indexed = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                  g_free, NULL);
    kozo_db_foreach (s->db, get_last_indexed, source);
    return TRUE;
}

BklSource *
bkl_source_removable_new (GMount *mount)
{
    BklSourceRemovable *removable;
    BklSource *source;
    GFile *root;
    char *uri;

    removable = g_new0 (BklSourceRemovable, 1);
    source = (BklSource *) removable;

    if (kozo_init (removable, mount) == FALSE) {
        g_free (source->name);
        g_free (removable);
        return NULL;
    }

    source->do_work = bkl_source_removable_do_work;
    source->investigate_files = bkl_source_removable_investigate_uris;
    source->add_item = bkl_source_removable_add_item;
    source->destroy = bkl_source_removable_destroy;

    root = g_mount_get_root (mount);
    uri = g_file_get_uri (root);
    g_object_unref (root);

    removable->finder = bkl_path_finder_new (source, uri);
    g_free (uri);

    return source;
}
