#include <libgupnp/gupnp-control-point.h>
#include <libgupnp/gupnp-device-info.h>

#include <bickley/bkl-db.h>
#include <bickley/bkl-entry.h>
#include <bickley/bkl-item-extended.h>

#include "bkl-finder.h"

typedef struct _BklSourceUPnP {
    BklSource source;

    BklFinder *finder;
    GUPnPDeviceInfo *info;
} BklSourceUPnP;

static void
bkl_source_upnp_destroy (BklSource *source)
{
    BklSourceUPnP *upnp = (BklSourceUPnP *) source;

    g_free (upnp);
}

static gboolean
bkl_source_upnp_do_work (BklSource *source)
{
    BklSourceUPnP *upnp = (BklSourceUPnP *) source;
    gboolean more_work;

    more_work = bkl_finder_lazy_dig (upnp->finder);
    return more_work;
}

static void
bkl_source_upnp_investigate_uris (BklSource *source,
                                  GPtrArray *files)
{
    BklSourceUPnP *upnp = (BklSourceUPnP *) source;
}

static GPtrArray *
copy_string_array (GPtrArray *original)
{
    GPtrArray *clone;
    int i;

    if (original == NULL) {
        return NULL;
    }

    clone = g_ptr_array_sized_new (original->len);
    for (i = 0; i < original->len; i++) {
        g_ptr_array_add (clone, g_strdup (original->pdata[i]));
    }

    return clone;
}

static gboolean
bkl_source_upnp_add_item (BklSource  *source,
                          const char *uri,
                          BklItem    *item,
                          GError    **error)
{
    BklSourceUPnP *upnp = (BklSourceUPnP *) source;
    BklItem *old;

    old = bkl_db_get_item (source->db, uri, error);
    if (*error != NULL) {
        if ((*error)->code != KOZO_DB_ERROR_KEY_NOT_FOUND) {
            return FALSE;
        }

        g_error_free (*error);
        *error = NULL;
    }

    if (old == NULL) {
        bkl_db_add_item (source->db, uri, item, error);
    } else {
        BklItemExtended *ext = (BklItemExtended *) old;
        GPtrArray *tags;

        /* Copy the extended details from the existing item
           as these shouldn't change on update */
        bkl_item_extended_set_use_count
            ((BklItemExtended *) item, bkl_item_extended_get_use_count (ext));
        bkl_item_extended_set_last_used
            ((BklItemExtended *) item, bkl_item_extended_get_last_used (ext));
        bkl_item_extended_set_rating
            ((BklItemExtended *) item, bkl_item_extended_get_rating (ext));
        bkl_item_extended_set_pinned
            ((BklItemExtended *) item, bkl_item_extended_get_pinned (ext));

        tags = bkl_item_extended_get_tags (ext);
        bkl_item_extended_set_tags ((BklItemExtended *) item,
                                    copy_string_array (tags));

        bkl_db_replace_item (source->db, uri, item, error);
        g_object_unref (old);
    }

    return TRUE;
}

static gboolean
kozo_init (BklSourceUPnP    *source)
{
    BklSource *s = (BklSource *) source;
    GError *error = NULL;
    const char *udn;
    char *name, *enc;

    udn = gupnp_device_info_get_udn (source->info);
    enc = g_compute_checksum_for_string (G_CHECKSUM_MD5, udn, -1);
    name = gupnp_device_info_get_friendly_name (source->info);

    s->db = bkl_db_get_for_name (enc, name, &error);
    g_free (enc);

    if (s->db == NULL) {
        g_warning ("Error getting Bickley database for %s: %s", udn,
                   error->message);
        g_free (name);
        return FALSE;
    }

    s->name = name;

    return TRUE;
}

BklSource *
bkl_source_upnp_new (GUPnPDeviceProxy *proxy)
{
    BklSourceUPnP *upnp;
    BklSource *source;

    upnp = g_new0 (BklSourceUPnP, 1);
    source = (BklSource *) upnp;

    upnp->info = (GUPnPDeviceInfo *) proxy;

    if (kozo_init (upnp) == FALSE) {
        g_free (source->name);
        g_free (upnp);
        return NULL;
    }

    source->do_work = bkl_source_upnp_do_work;
    source->investigate_files = bkl_source_upnp_investigate_uris;
    source->add_item = bkl_source_upnp_add_item;
    source->destroy = bkl_source_upnp_destroy;

    return source;
}
