/*  monkey-sound
 *  arch-tag: implementation of reading id3 tags over GIO
 *  Copyright (C) 2009 Intel Corporation
 *  Author: Iain Holmes <iain@linux.intel.com>
 *          based upon id3-vfs in Muine which is
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     based upon file.c in libid3tag by Robert Leslie
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "id3-gio.h"

# ifdef HAVE_CONFIG_H
#  include "config.h"
# endif

# include "global.h"

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# ifdef HAVE_ASSERT_H
#  include <assert.h>
# endif

# ifdef HAVE_UNISTD_H
#  include <unistd.h>
# endif

# include "file.h"
# include "tag.h"
# include "field.h"
# include "mp3bitrate.h"

struct giotag {
  struct id3_tag *tag;
  unsigned long location;
  id3_length_t length;
};

struct id3_gio_file {
  GFile *iofile;
  GInputStream *iostream;
  enum id3_gio_mode mode;
  int flags;
  int options;
  struct id3_tag *primary;
  unsigned int ntags;
  struct giotag *tags;
  goffset tag_offset;
};

enum {
  ID3_FILE_FLAG_ID3V1 = 0x0001
};

/*
 * NAME:	query_tag()
 * DESCRIPTION:	check for a tag at a file's current position
 */
static
signed long query_tag(GInputStream *iostream)
{
  goffset save_position;
  gssize bytes_read;
  id3_byte_t query[ID3_TAG_QUERYSIZE];
  signed long size;
  GError *error = NULL;

  save_position = g_seekable_tell((GSeekable *) iostream);

  bytes_read = g_input_stream_read(iostream, query, sizeof(query),
                                   NULL, &error);
  if (error != NULL) {
    g_warning ("Error reading from stream: %s", error->message);
    g_error_free (error);
    return 0;
  }

  size = id3_tag_query(query, bytes_read);

  g_seekable_seek((GSeekable *) iostream, save_position,
                  G_SEEK_SET, NULL, &error);
  if (error != NULL) {
    g_warning ("Error seeking stream: %s", error->message);
    g_error_free (error);
    return 0;
  }

  return size;
}

/*
 * NAME:	read_tag()
 * DESCRIPTION:	read and parse a tag at a file's current position
 */
static
struct id3_tag *read_tag(GInputStream *iostream, id3_length_t size)
{
  gssize bytes_read;
  id3_byte_t *data;
  struct id3_tag *tag = 0;
  GError *error = NULL;

  data = malloc(size);
  if (data) {
    bytes_read = g_input_stream_read (iostream, data, size, NULL, &error);
    if (error != NULL) {
      g_warning ("Error reading from stream: %s", error->message);
      g_error_free (error);

      free(data);
      return NULL;
    }

    tag = id3_tag_parse(data, bytes_read);

    free(data);
  }

  return tag;
}

/*
 * NAME:	update_primary()
 * DESCRIPTION:	update the primary tag with data from a new tag
 */
static
int update_primary(struct id3_tag *tag, struct id3_tag const *new)
{
  unsigned int i;

  if (!(new->extendedflags & ID3_TAG_EXTENDEDFLAG_TAGISANUPDATE))
    id3_tag_clearframes(tag);

  for (i = 0; i < new->nframes; ++i) {
    if (id3_tag_attachframe(tag, new->frames[i]) == -1)
      return -1;
  }

  return 0;
}

/*
 * NAME:	add_tag()
 * DESCRIPTION:	read, parse, and add a tag to a file structure
 */
static
int add_tag(struct id3_gio_file *file, id3_length_t length)
{
  goffset location;
  unsigned int i;
  struct giotag filetag, *tags;
  struct id3_tag *tag;

  location = g_seekable_tell ((GSeekable *) file->iostream);

  /* check for duplication/overlap */
  {
    unsigned long begin1, end1, begin2, end2;

    begin1 = location;
    end1   = begin1 + length;

    for (i = 0; i < file->ntags; ++i) {
      begin2 = file->tags[i].location;
      end2   = begin2 + file->tags[i].length;

      if (begin1 == begin2 && end1 == end2)
	return 0;  /* duplicate */

      if (begin1 < end2 && end1 > begin2)
	return -1;  /* overlap */
    }
  }

  tags = realloc(file->tags, (file->ntags + 1) * sizeof(*tags));
  if (tags == 0)
    return -1;

  file->tags = tags;

  tag = read_tag(file->iostream, length);
  if (tag == 0)
    return -1;

  if (update_primary(file->primary, tag) == -1) {
    id3_tag_delete(tag);
    return -1;
  }

  filetag.tag      = tag;
  filetag.location = location;
  filetag.length   = length;

  file->tags[file->ntags++] = filetag;

  id3_tag_addref(tag);

  return 0;
}

/*
 * NAME:	search_tags()
 * DESCRIPTION:	search for tags in a file
 */
static
int search_tags(struct id3_gio_file *file)
{
  goffset save_position;
  signed long size;
  int result = 0;
  GError *error = NULL;

  save_position = g_seekable_tell((GSeekable *)file->iostream);

  /* look for an ID3v1 tag */

  g_seekable_seek((GSeekable *) file->iostream, (goffset)-128, G_SEEK_END, NULL, &error);
  if (error != NULL) {
    g_warning ("Error seeking in stream: %s", error->message);
    g_error_free (error);
    goto fail;
  }

  size = query_tag(file->iostream);
  if (size > 0) {
    if (add_tag(file, size) == -1)
      goto fail;

    file->flags   |= ID3_FILE_FLAG_ID3V1;
    file->options |= ID3_FILE_OPTION_ID3V1;
  }

  /* look for a tag at the beginning of the file */
  g_seekable_seek((GSeekable *)file->iostream, 0, G_SEEK_SET, NULL, &error);
  if (error != NULL) {
    g_warning ("Error seeking in stream: %s", error->message);
    g_error_free (error);
    goto fail;
  }

  size = query_tag(file->iostream);
  if (size > 0) {
    struct id3_frame const *frame;

    if (add_tag(file, size) == -1)
      goto fail;

    /* locate tags indicated by SEEK frames */

    while ((frame =
	    id3_tag_findframe(file->tags[file->ntags - 1].tag, "SEEK", 0))) {
      long seek;

      seek = id3_field_getint(&frame->fields[0]);
      if (seek < 0 ||
          g_seekable_seek ((GSeekable *)file->iostream,
                           (goffset)seek, G_SEEK_CUR,
                           NULL, NULL) != TRUE)
	break;

      size = query_tag(file->iostream);
      if (size <= 0)
	break;
      else if (add_tag(file, size) == -1)
	goto fail;
    }
  }

  file->tag_offset = size;

  /* look for a tag at the end of the file (before any ID3v1 tag) */

  g_seekable_seek((GSeekable *) file->iostream,
                  ((file->flags & ID3_FILE_FLAG_ID3V1) ? -128 : 0) + -10,
                  G_SEEK_END, NULL, &error);
  if (error != NULL) {
    g_warning ("Error seeking in stream: %s", error->message);
    g_error_free (error);
    goto fail;
  }

  size = query_tag(file->iostream);
  if (size < 0 &&
      g_seekable_seek ((GSeekable *)file->iostream, (goffset)size, G_SEEK_CUR,
                       NULL, NULL)) {
    size = query_tag(file->iostream);
    if (size > 0 && add_tag(file, size) == -1)
      goto fail;
  }

  if (0) {
  fail:
    if (file->flags & ID3_FILE_FLAG_ID3V1)
	    result = 0;
    else
	   result = -1;
  }

  /* This really shouldn't be needed */
  save_position = g_seekable_tell((GSeekable *)file->iostream);

  return result;
}

/*
 * NAME:	finish_file()
 * DESCRIPTION:	release memory associated with a file
 */
static
void finish_file(struct id3_gio_file *file)
{
  unsigned int i;

  if (file->primary) {
    id3_tag_delref(file->primary);
    id3_tag_delete(file->primary);
  }

  for (i = 0; i < file->ntags; ++i) {
    id3_tag_delref(file->tags[i].tag);
    id3_tag_delete(file->tags[i].tag);
  }

  if (file->tags)
    free(file->tags);

  free(file);
}

/*
 * NAME:	new_file()
 * DESCRIPTION:	create a new file structure and load tags
 */
static
struct id3_gio_file *new_file(GFile *iofile, enum id3_gio_mode mode)
{
  struct id3_gio_file *file;
  GError *error = NULL;

  file = malloc(sizeof(*file));
  if (file == 0)
    goto fail;

  file->iofile = g_object_ref (iofile);
  file->iostream = (GInputStream *)g_file_read(iofile, NULL, &error);
  if (error != NULL) {
      g_warning ("Error opening %s for reading: %s", g_file_get_uri (iofile),
                 error->message);
    g_error_free (error);
    free (file);
    file = 0;

    goto fail;
  }

  file->mode    = mode;
  file->flags   = 0;
  file->options = 0;

  file->ntags   = 0;
  file->tags    = 0;

  file->tag_offset = 0;

  file->primary = id3_tag_new();
  if (file->primary == 0)
    goto fail;

  id3_tag_addref(file->primary);

  /* load tags from the file */

  if (search_tags(file) == -1)
    goto fail;

  if (0) {
  fail:
    if (file) {
      finish_file(file);
      file = NULL;
    }
  }

  return file;
}

/*
 * NAME:	file->open()
 * DESCRIPTION:	open a file given its pathname
 */
struct id3_gio_file *id3_gio_open(GFile *iofile, enum id3_gio_mode mode)
{
  struct id3_gio_file *file;

  file = new_file(iofile, mode);
  if (file == 0)
    g_object_unref(iofile);

  return file;
}

/*
 * NAME:	file->fdopen()
 * DESCRIPTION:	open a file using an existing file descriptor
 *
 * Unsupported with the gio interface
 */

/*
 * NAME:	file->close()
 * DESCRIPTION:	close a file and delete its associated tags
 */
void id3_gio_close(struct id3_gio_file *file)
{
  g_input_stream_close(file->iostream, NULL, NULL);
  g_object_unref(file->iostream);
  g_object_unref(file->iofile);
  finish_file(file);
}

/*
 * NAME:	file->tag()
 * DESCRIPTION:	return the primary tag structure for a file
 */
struct id3_tag *id3_gio_tag(struct id3_gio_file const *file)
{
  return file->primary;
}

/*
 * NAME:	file->update()
 * DESCRIPTION:	rewrite tag(s) to a file
 */
int id3_gio_update(struct id3_gio_file *file)
{
  id3_length_t size;
  unsigned char id3v1_data[128], *id3v1 = 0, *id3v2 = 0;

  if (file->mode != ID3_GIO_MODE_READWRITE)
    return -1;

  if (file->options & ID3_FILE_OPTION_ID3V1) {
    file->primary->options |= ID3_TAG_OPTION_ID3V1;

    size = id3_tag_render(file->primary, 0);
    if (size) {
      assert(size == sizeof(id3v1_data));

      size = id3_tag_render(file->primary, id3v1_data);
      if (size) {
	assert(size == sizeof(id3v1_data));
	id3v1 = id3v1_data;
      }
    }
  }

  file->primary->options &= ~ID3_TAG_OPTION_ID3V1;

  size = id3_tag_render(file->primary, 0);
  if (size) {
    id3v2 = malloc(size);
    if (id3v2 == 0)
      return -1;

    size = id3_tag_render(file->primary, id3v2);
    if (size == 0) {
      free(id3v2);
      id3v2 = 0;
    }
  }

  /* ... */

  if (id3v2)
    free(id3v2);

  return 0;
}

static int
id3_gio_is_wave (guchar *buffer)
{
  if (buffer[8] != 'W')
    return 0;
  if (buffer[9] != 'A')
    return 0;
  if (buffer[10] != 'V')
    return 0;
  if (buffer[11] != 'E' && buffer[11] != ' ')
    return 0;

  return 1;
}

int
id3_gio_bitrate (struct id3_gio_file *file, int *bitrate, int *samplerate,
		 int *time, int *version, int *vbr, int *channels)
{
  goffset save_position;
  gssize length_read;
  GInputStream *iostream = file->iostream;
  guchar buffer[16384];
  int is_wave, found, i;
  GError *error = NULL;

  *bitrate = 0;
  *samplerate = 0;
  *time = 0;
  *channels = 0;
  *version = 0;
  *vbr = 0;
  found = 0;

  save_position = g_seekable_tell((GSeekable *) iostream);

  g_seekable_seek ((GSeekable *) iostream, file->tag_offset, G_SEEK_SET,
                   NULL, NULL);

  length_read = g_input_stream_read (iostream, buffer, sizeof (buffer),
                                     NULL, &error);
  if (error != NULL) {
    g_warning ("Error reading from stream: %s", error->message);
    g_error_free (error);
    goto bitdone;
  }

  if(length_read < 512)
    goto bitdone;

  /* Reduce false positive by castrating the search if we have a WAVE file */
  is_wave = id3_gio_is_wave (buffer);
  if (is_wave == 1)
    length_read = 4096;

  for (i = 0; i + 4 < length_read; i++)
  {
    if (mp3_bitrate_parse_header (buffer+i, length_read - i, bitrate, samplerate, time, version, vbr, channels))
    {
      found = 1;
      break;
    }
  }

bitdone:
  g_seekable_seek((GSeekable *) iostream, save_position, G_SEEK_SET, NULL, NULL);

  return found;
}

