/*  monkey-sound
 *
 *  arch-tag: Implementation of various Ogg Vorbis loading functions
 *
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Modified by Iain Holmes <iain@linux.intel.com>
 * Copyright (C) 2008 Intel Corporation.
 */

#include <gio/gio.h>
#include "ogg-helper.h"

size_t
ogg_helper_read (void *ptr, size_t size, size_t nmemb, void *datasource)
{
	gssize bytes_read;
	GInputStream *iostream = (GInputStream *)datasource;
        GError *error = NULL;

	bytes_read = g_input_stream_read(iostream, ptr, (gssize) size * nmemb,
                                         NULL, &error);
        if (error != NULL) {
            g_warning ("Error reading stream: %s", error->message);
            g_error_free (error);
            return -1;
	}

	return (size_t) bytes_read;
}

int
ogg_helper_seek (void *datasource, ogg_int64_t offset, int whence)
{
	GSeekable *seekable = (GSeekable *)datasource;
        GSeekType mode;
        GError *error = NULL;

        /* GSeekType does not match up with the fseek definitions that
           @whence uses */
        switch (whence) {
        case SEEK_SET:
            mode = G_SEEK_SET;
            break;

        case SEEK_CUR:
            mode = G_SEEK_CUR;
            break;

        case SEEK_END:
            mode = G_SEEK_END;
            break;

        default:
            mode = G_SEEK_SET;
            break;
        }

	g_seekable_seek(seekable, offset, mode, NULL, &error);
        if (error != NULL) {
            g_warning ("Error seeking: %s", error->message);
            g_error_free (error);
            return -1;
        }

	return 0;
}

int
ogg_helper_close (void *datasource)
{
	GInputStream *iostream = (GInputStream *)datasource;

        g_input_stream_close (iostream, NULL, NULL);
        g_object_unref (iostream);

	return 0;
}

/*
 * VorbisFile behaves inconsistently when it comes to close or not files
 * <xiphmont> If it's *not* a vorbis file, it doesn't close it.  If it
 * *is* but there's an error it does.  I need to correct that
 * <xiphmont> Like I said, this inconsistency only pops up on filesystem
 * fault.  Or a buggy callback ;-)
 *
 * Proposed work-around in applications:
 * <xiphmont> make the 'close' callback do nothing and close it yourself is
 * one way.
 *
 * Thanks to Monty for the help
 */
int
ogg_helper_close_dummy (void *datasource)
{
	return 0;
}

long
ogg_helper_tell (void *datasource)
{
	GSeekable *seekable = (GSeekable *)datasource;

	return (long) g_seekable_tell (seekable);
}

